##############################################################################
#
#    Accounting periods, for Odoo
#    Copyright (C) 2018 XCG Consulting <http://odoo.consulting>
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation, either version 3 of the
#    License, or (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
##############################################################################
import logging
from dateutil.relativedelta import relativedelta

from odoo import fields
from odoo import models
from odoo import api
from odoo.exceptions import UserError, ValidationError
from odoo import _
from odoo import SUPERUSER_ID

_logger = logging.getLogger(__name__)


class AccountFiscalyear(models.Model):
    """Account fiscal year defined an accounting year.
    """

    _name = 'account.fiscalyear'
    _description = 'Fiscal year'

    _order = 'code'

    code = fields.Char(
        string='Code',
        help='Unique code of this fiscal year.',
        index=True,
        required=True,
    )

    company_id = fields.Many2one(
        comodel_name='res.company',
        string='Company',
        default=lambda rec: rec.env.user.company_id,
        help='The company the fiscal year is in.',
        required=True,
    )

    name = fields.Char(
        string='Name',
        help='Displayed name of this fiscal year.',
        required=True,
    )

    period_ids = fields.One2many(
        comodel_name='account.period',
        inverse_name='fiscalyear_id',
        string='Periods',
        help='The periods inside this fiscal year.',
    )

    end_date = fields.Date(
        string='End date',
        help=(
            'When this period ends'
        ),
        required="True",
    )

    _sql_constraints = [
        ('unique_code_per_company',
         'UNIQUE(code, company_id)',
         'The code must be unique.'),
    ]

    @api.model
    def _check_end_date(self):
        """ Check end date value when fiscal
        year last day/month is provided.
        """
        config_fy_month = self.env.user.company_id.fiscalyear_last_month
        config_fy_day = self.env.user.company_id.fiscalyear_last_day
        if not config_fy_day and not config_fy_month:
            return

        if not self.end_date:
            raise UserError(_(
                "Generate periods need to set end date."
            ))

        ed = fields.Date.from_string(self.end_date)
        if (
            ed.month > config_fy_month or
            ed.month == config_fy_month and
            ed.day > config_fy_day
        ):
            raise UserError(_(
                "The end date must match the fiscal "
                "year general config (%s/%s)"
            ) % (config_fy_month, config_fy_day))

    @api.multi
    def create_period(self):
        """ Call by the "Create Periods" Button.

            This function use "end date" field to generate
            automatically all periods.
        """
        for fy in self:
            fy._check_end_date()

            ed = fields.Date.from_string(fy.end_date)
            sd = ed - relativedelta(months=12, days=-1)

            while sd < ed:
                period_ed = sd + relativedelta(months=1, days=-1)
                if fields.Date.to_string(period_ed) > fy.end_date:
                    period_ed = fields.Date.from_string(fy.end_date)

                fy.write({
                    'period_ids': [
                        (0, 0, {
                            'name': sd.strftime('%m/%Y'),
                            'code': sd.strftime('%m/%Y'),
                            'date_start': sd,
                            'date_stop': period_ed
                        })
                    ]
                })
                sd = sd + relativedelta(months=1)

        return True

    @api.multi
    def action_realloc_period_am(self):
        """ Call by the "Reallocate Periods" action.

        Look up all accounts move where periods is empty and
        try to reallocate it.
        """
        for record in self:
            # Only administrator can reallocate periods.
            if not record.env.uid == SUPERUSER_ID:
                raise ValidationError(_(
                    'It not possible to run this action,'
                    'please contact your administrator.'
                ))

            account_moves = record.env['account.move'].search([
                ('period_id', '=', False),
                ('state', '=', 'posted'),
            ])

            for am in account_moves:
                # Cache some data.
                company = am.company_id

                # Periods are ordered by date so selecting the first one is
                # fine.
                period = record.env['account.period'].search([
                    ('company_id', '=', company.id),
                    ('date_start', '<=', am.date),
                    ('date_stop', '>=', am.date),
                ], limit=1)

                if period:
                    _logger.info(
                        "reallocate period :%s: onto account move :%s:",
                        period.code, am.name)

                    am.write({
                        'period_id': period.id,
                        'transaction_date': am.date
                    })
